
import argparse
import os

from winservicetools import utils

def cli_main():
	
    script_pargs = ('--script',)
    script_kwargs = {
                     'dest': 'script_path',
                     'required': False,
                     'type': os.path.abspath,
                    }
    package_pargs = ('--import',)
    package_kwargs = {
                      'dest': 'import_name',
                      'required': False,
                     }
    
    desc = 'Install Python services via sc.exe'
    parser = argparse.ArgumentParser(description=desc)
    subparsers = parser.add_subparsers()
    
    # Installer subparser
    parser_install = subparsers.add_parser('install', help='Install services')
    parser_install.set_defaults(func=utils.install)
    group = parser_install.add_mutually_exclusive_group(required=True)
    ts_kwargs = script_kwargs.copy()
    ts_kwargs['help'] = 'Path to script to run as service'
    tp_kwargs = package_kwargs.copy()
    tp_kwargs['help'] = 'Package import to run as service'
    group.add_argument(*script_pargs, **ts_kwargs)
    group.add_argument(*package_pargs, **tp_kwargs)
    
    # Delete subparser
    parser_delete = subparsers.add_parser('delete', help='Delete services')
    parser_delete.set_defaults(func=utils.delete)
    group = parser_delete.add_mutually_exclusive_group(required=True)
    ts_kwargs = script_kwargs.copy()
    ts_kwargs['help'] = 'Path to service script to delete'
    tp_kwargs = package_kwargs.copy()
    tp_kwargs['help'] = 'Package import to delete service'
    group.add_argument(*script_pargs, **ts_kwargs)
    group.add_argument(*package_pargs, **tp_kwargs)
    
    # Update subparser
    parser_update = subparsers.add_parser('update', help='Update services')
    parser_update.set_defaults(func=utils.update)
    group = parser_update.add_mutually_exclusive_group(required=True)
    ts_kwargs = script_kwargs.copy()
    ts_kwargs['help'] = 'Path to service script to update'
    tp_kwargs = package_kwargs.copy()
    tp_kwargs['help'] = 'Package import to update service'
    group.add_argument(*script_pargs, **ts_kwargs)
    group.add_argument(*package_pargs, **tp_kwargs)

    cli_args = parser.parse_args()
    
    kwargs = vars(cli_args)
    func = kwargs.pop('func')
    func(**kwargs)
    




import ctypes
import importlib
import os
import subprocess
import sys

from winservicetools import WindowsSvc

def find_service_cls(module, cls_name=None):
    """ Dark magic to find the service class

        This function takes a module object and then searches through
        it looking for a object that has a type that is type, indicating
        that it is a class and then is also a subclass of the primary
        WindowsSvc class. It returns the first item that meets this
        criteria.

        cls_name allows you to pass a cls_name that is also checked in
        case there are multiple WindowSvc classes in the same file.

    """
    for entry in dir(module):
        obj = getattr(module, entry)
        if type(obj) is type and issubclass(obj, WindowsSvc):
            if cls_name is None:
                return obj
            if cls_name == obj.__name__:
                return obj
    if cls_name:
        msg = 'No service class found with name "{}"'
        err = ValueError(msg.format(cls_name))
    else:
        msg = 'No service class found in module {}'.format(module.__name__)
        err = ValueError(msg)
    raise err

def load_module_from_path(path):
    """ Loads a module at the given path

        This code is basically lifted from the Python docs and loads
        a module dynamically from a path. This is used for the scripts
        to load the service class.
    """

    module_name = "service_module"
    spec = importlib.util.spec_from_file_location(module_name, path)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)

    return module

def get_sc_args(cls, inputstring):
    """ Return formatted arguments for sc.exe description and config """

    exe_name = cls._exe_name_ if cls._exe_name_ else sys.executable
    exe_args = cls._exe_args_ if cls._exe_args_ else tuple()
    svc_binpath = ' '.join((exe_name, *exe_args, inputstring))

    svc_name = cls._svc_name_
    # I don't understand exactly what is going on here but if you read the
    # sc.exe documentation it is very explicit that there must be a space
    # after the equals sign for each option. However when including this
    # space here subprocess does something funky and it is interpreted
    # a a literal space in front of each value. This isn't really and issue
    # except for binPath which puts a space in front of the command to execute
    # the service and will readily cause the service to fail to
    # start. So counter to official Windows documentation don't put a
    # space between the option and the value.
    binpath = 'binPath={}'.format(svc_binpath)
    displayname = 'DisplayName={}'.format(cls._svc_display_name_)

    config = [displayname]

    if cls._svc_start_:
        config.append('start={}'.format(cls._svc_start_))
    if cls._svc_depends_:
        dependslist = '/'.join(cls._svc_depends)
        config.append('depend={}'.format(dependslist))
    if cls._svc_account_:
        config.append('obj={}'.format(cls._svc_account_))
    if cls._svc_password_:
        config.append('password={}'.format(cls._svc_password_))
    if cls._svc_group_:
        config.append('group={}'.format(cls._svc_group_))

    out = {
           'name': svc_name,
           'binpath': binpath,
           'description': cls._svc_description_,
           'config': config,
          }
    return out

def process_inputs(script_path=None, import_name=None, cls_name=None):
    """ Adjudicate inputs and return the class and input string

        This function just determines if a script or import was passed
        and then ensures the correct install string is given and retrieves
        the class.

    """

    if script_path is None and import_name is None:
        raise ValueError('Must provide script path or import path')
    elif script_path and import_name:
        msg  = (
                'Please provide either a script path or a import'
                ' path not both'
               )
        raise ValueError(msg)
    elif script_path:
        # Check to make sure the script path exists because if it doesn't
        # we get a very cryptic error message.
        if not os.path.isfile(script_path):
            msg = 'Script path {} does not exist'.format(script_path)
            raise ValueError(msg)
        module = load_module_from_path(script_path)
        inputstring = script_path
    elif import_name:
        module = importlib.import_module(import_name)
        inputstring = '-m {}'.format(import_name)
    else:
        msg = (
               'There was an issue interpreting the script or import'
               ' path arguments'
              )
        raise ValueError(msg)

    cls = find_service_cls(module, cls_name)

    return cls, inputstring

def install(script_path=None, import_name=None, cls_name=None):
    """ Install and configure service using subprocess and sc.exe """

    if ctypes.windll.shell32.IsUserAnAdmin() == 0:
        raise RuntimeError('Must be admin to install services')

    cls, inputstring = process_inputs(script_path, import_name, cls_name)
    sc_args = get_sc_args(cls, inputstring)
    svc_name = sc_args['name']

    # CMD 1 creates the service
    # CMD 2 updates all the parameters of the service some of this could be
    # done during creation but I think this approach is cleaner
    # CMD 3 just sets the description which has to be done separate for some
    # reason.
    cmd1 = ('sc.exe', 'create', svc_name, sc_args['binpath'])
    cmd2 = ('sc.exe', 'config', svc_name) + tuple(sc_args['config'])
    cmd3 = ('sc.exe', 'description', svc_name, sc_args['description'])

    # We run each sub process and then check the return value and raise and
    # error if it isn't zero.
    cp = subprocess.run(cmd1)
    if cp.returncode != 0:
        raise RuntimeError('Service failed to install')

    cp = subprocess.run(cmd2)
    if cp.returncode != 0:
        raise RuntimeError('Failed to configure service')

    cp = subprocess.run(cmd3)
    if cp.returncode != 0:
        raise Warning('Service description was not set')

def delete(script_path=None, import_name=None, cls_name=None):
    """ Delete service using subprocess and sc.exe delete """

    if ctypes.windll.shell32.IsUserAnAdmin() == 0:
        raise RuntimeError('Must be admin to delete services')

    cls, inputstring = process_inputs(script_path, import_name, cls_name)
    svc_name = cls._svc_name_

    cmd1 = ('sc.exe', 'query', svc_name)
    cmd2 = ('sc.exe', 'stop', svc_name)
    cmd3 = ('sc.exe', 'delete', svc_name)

    cp = subprocess.run(cmd1, capture_output=True)
    if cp.returncode != 0:
        raise RuntimeError('Service query failed')

    if b'STOPPED' not in cp.stdout:
        cp = subprocess.run(cmd2)
        if cp.returncode != 0:
            raise RuntimeError('Unable to stop service')

    cp = subprocess.run(cmd3)
    if cp.returncode != 0:
        raise RuntimeError('Service not deleted')

def update(script_path=None, import_name=None, cls_name=None):
    """ Update service configuration using sc.exe config """

    if ctypes.windll.shell32.IsUserAnAdmin() == 0:
        raise RuntimeError('Must be admin to update services')
    
    cls, inputstring = process_inputs(script_path, import_name, cls_name)
    sc_args = get_sc_args(cls, inputstring)
    svc_name = sc_args['name']

    configargs = (sc_args['binpath'],) + tuple(sc_args['config'])
    cmd1 = ('sc.exe', 'config', svc_name) + configargs
    cmd2 = ('sc.exe', 'description', svc_name, sc_args['description'])

    cp = subprocess.run(cmd1)
    if cp.returncode != 0:
        raise RuntimeError('Failed to configure service')

    cp = subprocess.run(cmd2)
    if cp.returncode != 0:
        raise RuntimeError('Failed to update service description')




import logging
import threading
import traceback

import win32serviceutil
import win32service
import win32event
import servicemanager

from .__version__ import __version__

logger = logging.getLogger(__name__)

class WindowsSvc(win32serviceutil.ServiceFramework):

    def __init__(self, args):
        super().__init__(args)
        self.hWaitStop = win32event.CreateEvent(None,0,0,None)
        self.service_event = threading.Event()

    def SvcStop(self):
        self.ReportServiceStatus(win32service.SERVICE_STOP_PENDING)
        logger.info('Stop event received')
        logger.debug('Clearing run event')
        self.service_event.clear()
        logger.debug('Run event cleared')
        # This call blocks so you need to run any shutdown code before
        # calling this
        win32event.SetEvent(self.hWaitStop)

    def SvcDoRun(self):
        logger.info('Service target starting')
        pargs = (
                 servicemanager.EVENTLOG_INFORMATION_TYPE,
                 servicemanager.PYS_SERVICE_STARTED,
                 (self._svc_name_,''),
                )
        servicemanager.LogMsg(*pargs)
        try:
            logger.debug('Set service event')
            self.service_event.set()
            logger.debug('Service event is set')
            logger.debug('Start target')
            logger.debug('Target positional arguments: %s', self.target_args)
            logger.debug('Target keyword arguments: %s', self.target_kwargs)
            self.target(*self.target_args, service_event=self.service_event, **self.target_kwargs)
        except Exception:
            # Catch all exceptions and log to both the Python logger
            # and log it to the Windows event log.
            logger.exception('Exception in target')
            servicemanager.LogErrorMsg(traceback.format_exc())

        pargs = (
                 servicemanager.EVENTLOG_INFORMATION_TYPE,
                 servicemanager.PYS_SERVICE_STOPPED,
                 (self._svc_name_, ''),
                )
        servicemanager.LogMsg(*pargs)
        logger.info('Target has exited service is shutting down')

    @classmethod
    def start(cls):
        """ Starts service """

        logger.info('Python windows service %s starting', cls.__name__)
        servicemanager.Initialize()
        servicemanager.PrepareToHostSingle(cls)
        servicemanager.StartServiceCtrlDispatcher()
        logger.info('Python Windows Service %s stopped', cls.__name__)

    @classmethod
    def new_service(cls, target, svc_name, svc_display_name,
                    svc_description=None, svc_depends=None, svc_start='auto',
                    svc_account=None, svc_password=None, svc_group=None,
                    exe_name=None, exe_args=None, cls_name=None, target_args=(),
                    target_kwargs=None):
        """ Generate new generic Windows service class """

        kwargs = {}

        # Setup Windows service arguments
        kwargs['_svc_name_'] = svc_name
        kwargs['_svc_display_name_'] = svc_display_name
        kwargs['_svc_description_'] = svc_description
        kwargs['_svc_depends_'] = svc_depends
        kwargs['_svc_start_'] = svc_start
        kwargs['_svc_account_'] = svc_account
        kwargs['_svc_password_'] = svc_password
        kwargs['_svc_group_'] = svc_group
        kwargs['_exe_name_'] = exe_name
        kwargs['_exe_args_'] = exe_args

        # Setup target and target arguments
        # We call static method here on target other wise it will get
        # passed the class implicitly
        kwargs['target'] = staticmethod(target)
        kwargs['target_args'] = target_args
        kwargs['target_kwargs'] = target_kwargs if target_kwargs else {}

        # If class name isn't give just use the service name
        if cls_name is None:
            cls_name = svc_name

        # Make a call to type to generate a new object with the given input
        # All the keyword arguments passed will be
        obj = type(cls_name, (cls,), kwargs)

        return obj

    @classmethod
    def install(cls, scriptpath=None, modulepath=None):

        import sys
        import subprocess

        if scriptpath is None and modulepath is None:
            raise ValueError('Must provide script path or module path')
        elif scriptpath and modulepath:
            msg  = (
                    'Please provide either a script path or a module'
                    ' path not both'
                   )
            raise ValueError(msg)
        elif scriptpath:
            inputstring = scriptpath
        elif modulepath:
            inputstring = '-m {}'.format(modulepath)
        else:
            msg = (
                   'There was an issue interpeting the script and module'
                   ' path arguments'
                  )
            raise ValueError(msg)

        pythonexe = sys.executable
        svc_execution_str = "{pythonexe:} {inputstring:}"
        kwargs = {
                  'pythonexe': pythonexe,
                  'inputstring': inputstring
                 }
        svc_execution_cmd = svc_execution_str.format(**kwargs)

        svc_name = cls._svc_name_
        binpath = 'binPath= {}'.format(svc_execution_cmd)
        displayname = 'DisplayName= {}'.format(cls._svc_display_name_)

        # CMD 1 creates the service and CMD2 adds the description
        # you can't add a description until the command is created
        cmd1 = (
                'sc.exe', 'create', svc_name,
                binpath,
                displayname,
                # 'start= auto'
               )
        cmd2 = ('sc.exe', 'descrption', svc_name, cls._svc_description_)

        cp = subprocess.run(cmd1)
        if cp.returncode != 0:
            raise RuntimeError('Service failed to install')

        cp = subprocess.run(cmd2)
        if cp.returncode != 0:
            raise Warning('Service description was not set')

    @classmethod
    def delete(cls):

        import subprocess

        cmd1 = ('sc.exe', 'query', cls._svc_name_)
        cmd2 = ('sc.exe', 'stop', cls._svc_name_)
        cmd3 = ('sc.exe', 'delete', cls._svc_name_)

        cp = subprocess.run(cmd1, capture_output=True)
        if cp.returncode != 0:
            raise RuntimeError('Service query failed')
        breakpoint()

        cp = subprocess.run(cmd2)
        if cp.returncode != 0:
            raise RuntimeError('Service not deleted')



if __name__ == '__main__':

    kwargs = {}
    kwargs['svc_name'] = 'testservice'
    kwargs['svc_display_name'] = 'testservice'
    kwargs['svc_description'] = 'Python test service'
    kwargs['svc_depends'] = None
    kwargs['exe_name'] = None
    kwargs['exe_args'] = None
    testservice = WindowsSvc.new_service(**kwargs)